# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import sys
import obob_condor
import mne
import os
import mne.commands.mne_make_scalp_surfaces
from .utils.jobs import JobWithFreesurfer


class MakeSourceSpacesJob(JobWithFreesurfer):
    def run(self, subject_id):
        super(MakeSourceSpacesJob, self).run()
        src = mne.setup_source_space(subject_id,
                                     subjects_dir=self.subjects_dir,
                                     n_jobs=1)

        final_outpath = os.path.join(self.subjects_dir, subject_id, 'bem')
        mne.write_source_spaces(
            os.path.join(final_outpath, '%s-oct6-src.fif' % (subject_id,)),
            src, overwrite=True)

        bem_fname = os.path.join(final_outpath,
                                 '%s-complete-bem.fif' % (subject_id,))

        fs_avg_mri_path = os.path.join(self.subjects_dir, 'fsaverage', 'mri',
                                       'T1.mgz')
        vol_src = mne.setup_volume_source_space(subject_id,
                                                subjects_dir=self.subjects_dir,
                                                mri=fs_avg_mri_path,
                                                bem=bem_fname)

        mne.write_source_spaces(
            os.path.join(final_outpath, '%s-5mm-vol-src.fif' % (subject_id,)),
            vol_src, overwrite=True)

        vol_src = mne.setup_volume_source_space(subject_id,
                                                subjects_dir=self.subjects_dir,
                                                mri=fs_avg_mri_path,
                                                bem=bem_fname,
                                                pos=10)

        mne.write_source_spaces(
            os.path.join(final_outpath, '%s-10mm-vol-src.fif' % (subject_id,)),
            vol_src, overwrite=True)


def make_freesurfer_sourcespaces(subject_id):
    os.chdir('/mnt/obob/tmp')

    job_cluster = obob_condor.JobCluster(working_directory='/mnt/obob/tmp')
    job_cluster.add_job(MakeSourceSpacesJob, subject_id)

    job_cluster.submit()


def make_freesurfer_sourcespaces_cmd():
    subject_id = sys.argv[1]
    make_freesurfer_sourcespaces(subject_id)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import sys
import subprocess
import obob_condor
import mne
import os
import mne.commands.mne_make_scalp_surfaces
from .utils.jobs import JobWithFreesurfer


class FreesurferBEMJob(JobWithFreesurfer):
    freesurfer_init = 'export FREESURFER_HOME=/mnt/obob/bin/freesurfer; ' \
                      'export SUBJECTS_DIR=/mnt/obob/freesurfer_subjects; ' \
                      'source $FREESURFER_HOME/SetUpFreeSurfer.sh; '

    watershed_prefloods = [25, 20, 30, 15, 35, 10, 40, 5, 45, 2, 50, 1, 55, 65,
                           70, 75, 80, 85, 90]

    extra_paths = ['/mnt/obob/bin/freesurfer/bin',
                   '/mnt/obob/bin/freesurfer/fsfast/bin',
                   '/mnt/obob/bin/freesurfer/tktools',
                   '/mnt/obob/bin/freesurfer/mni/bin']

    def run(self, subject_id):
        super(FreesurferBEMJob, self).run()
        subjects_dir = '/mnt/obob/freesurfer_subjects'

        is_running_file = os.path.join(subjects_dir, subject_id, 'scripts',
                                       'IsRunning.lh+rh')
        if os.path.exists(is_running_file):
            os.unlink(is_running_file)
        subprocess.check_call(
            [self.freesurfer_init +
             ' recon-all -make all -no-isrunning -s ' +
             subject_id], shell=True)

        final_outpath = os.path.join(subjects_dir, subject_id, 'bem')

        for cur_preflood in self.watershed_prefloods:
            mne.bem.make_watershed_bem(subject=subject_id,
                                       subjects_dir=subjects_dir, atlas='T1',
                                       overwrite=True, preflood=cur_preflood)

            with mne.utils.ArgvSetter(
                    ('-s', subject_id, '-n', '--force', '--overwrite')):
                mne.commands.mne_make_scalp_surfaces.run()

            try:
                bem = mne.make_bem_model(subject_id, subjects_dir=subjects_dir)
            except RuntimeError:
                pass
            else:
                break

        bem_fname = os.path.join(final_outpath,
                                 '%s-complete-bem.fif' % (subject_id,))
        mne.write_bem_surfaces(bem_fname, bem)

        bem_solution = mne.make_bem_solution(bem)
        bem_solution_fname = os.path.join(
            final_outpath, '%s-complete-bem-sol.fif' % (subject_id,))
        mne.write_bem_solution(bem_solution_fname, bem_solution)


def make_freesurfer_bem(subject_id):
    os.chdir('/mnt/obob/tmp')

    job_cluster = obob_condor.JobCluster(working_directory='/mnt/obob/tmp')
    job_cluster.add_job(FreesurferBEMJob, subject_id)

    job_cluster.submit()


def make_freesurfer_bem_cmd():
    subject_id = sys.argv[1]
    make_freesurfer_bem(subject_id)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
# obob_mne is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# obob_mne is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with obob_mne. If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def _get_Y_from_metadata(data, query_list):
    n_epochs = len(data)

    y = np.zeros(n_epochs)
    my_metadata = data.metadata.reset_index()

    for idx, cur_query in enumerate(query_list):
        epochs_idx = my_metadata.query(cur_query).index
        y[epochs_idx] = idx + 1

    if np.any(y == 0):
        raise RuntimeError('Not all of the epochs were selected!')

    return y



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import os
import zipfile
import shutil
import subprocess
import sys
import tempfile
import glob


def import_subject():
    full_file_name = sys.argv[1]
    f_name = os.path.basename(full_file_name)
    subject_id = f_name[0:12]

    tmp_folder = tempfile.mkdtemp()
    with zipfile.ZipFile(full_file_name, 'r') as mri_zip:
        mri_zip.extractall(tmp_folder)

    first_file = glob.glob(os.path.join(tmp_folder, '*'))[0]
    subprocess.call(['source /etc/profile.d/zz_activate_freesurfer.sh; '
                     'recon-all -i ' + first_file + ' -s ' + subject_id],
                    shell=True)

    shutil.rmtree(tmp_folder)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import mne
import obob_mne.events
import scipy


def collapse_conditions(epochs, keep_condition):
    """Merge conditions in event_id.

    Supposed we have an event_id like this:

    ``
    event_id = ['attention:audio/stim:audio',
                'attention:visual/stim:audio',
                'attention:audio/stim:visual',
                'attention:visual/stim:visual']
    ``

    And we want to compare the two stim conditions, ignoring the attention
    factor. Then we use this function:

    .. code-block:: python

        event_id = collapse_conditions(epochs, 'stim')

    Parameters
    ----------
    epochs : :class:`mne.Epochs`
        The  :class:`mne.Epochs` to modify
    keep_condition : str
        The factor to keep

    Returns
    -------
    epochs : :class:`mne.Epochs`
        The modified instance

    """
    epochs = epochs.copy()
    all_factors = list()
    collapsed_factors = list()
    for cur_key in epochs.event_id.keys():
        all_factors += [x for x in cur_key.split('/') if keep_condition in x]
        collapsed_factors += [x for x in cur_key.split('/') if
                              keep_condition not in x]

    all_factors = set(all_factors)

    new_event_id = dict()
    for idx, cur_factor in enumerate(all_factors):
        this_id = idx
        while this_id in epochs.events[:, 2]:
            this_id += 1

        all_values = obob_mne.events.filter_event_id(epochs.event_id,
                                                     cur_factor).values()

        epochs.events = mne.merge_events(events=epochs.events,
                                         ids=all_values,
                                         new_id=this_id,
                                         replace_events=True)
        new_event_id[cur_factor] = this_id

    epochs.event_id = new_event_id
    epochs.info['collapsed_factors'] = '/'.join(set(collapsed_factors))

    return epochs


def diff_collapsed_conditions(epochs_list):
    """No idea."""
    all_collapsed_factors = [x.info['collapsed_factors'] for x in epochs_list]

    for cur_epochs in epochs_list:
        cur_collapsed = cur_epochs.info['collapsed_factors']
        cur_coll_set = set(cur_collapsed.split('/'))

        for cur_cmp_epoch, cur_original_factors in zip(epochs_list,
                                                       all_collapsed_factors):
            if not cur_epochs == cur_cmp_epoch:
                cur_cmp_set = set(cur_original_factors.split('/'))
                cur_coll_set -= cur_cmp_set

        cur_epochs.info['collapsed_factors'] = '/'.join(cur_coll_set)


def no_diff_collapsed_conditions(epochs_list):
    """No idea."""
    all_collapsed_factors = [x.info['collapsed_factors'] for x in epochs_list]
    collapsed_factors_set = [set(x.split('/')) for x in all_collapsed_factors]

    final_set = collapsed_factors_set[0]

    for cur_set in collapsed_factors_set[1:]:
        final_set -= cur_set


def zscore_epochs(epoch):
    data = epoch.get_data()
    data = data.swapaxes(0, 1)

    original_shape = data.shape

    data = data.reshape((data.shape[0], -1))
    data = scipy.stats.zscore(data, 1)

    data = data.reshape(original_shape)
    data = data.swapaxes(0, 1)

    epoch._data = data

    return epoch



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
# obob_mne is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# obob_mne is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with obob_mne. If not, see <http://www.gnu.org/licenses/>.

import mne
import numpy
import copy
import logging
import collections
import scipy.stats

from .temporal import TemporalArray
from .helpers import _get_Y_from_metadata


class GeneralizedTemporalArray(mne.utils.SizeMixin):
    """Base class for Temporal Generalization Decoding.

    Parameters
    ----------
    scores_raw : :class:`numpy.ndarray` shape (n_times, n_times) or \
                (n_folds, n_times, n_times)
        Classifier accuracies.
    scoring_name : str
        Name of the scoring function (i.e. Accuracy...)
    n_classes : int
        Number of classes of the classification
    c_factors_training : str
        Name of the factors over which was collapsed in the training set
    weights : :class:`numpy.ndarray` shape (n_channels, n_times)
        Classifier weights
    info : dict
        Info dict
    times_training : :class:`numpy.ndarray` shape (n_times)
        Array of times in seconds of the training data
    tmin : int
        ???
    times_testing : None or :class:`numpy.ndarray` shape (n_times)
        Array of times in seconds of the testing data. If ``None``,
        it is copied from times_training.
    c_factors_testing : None or str
        Name of the factors over which was collapsed in the testing set.
        If ``None``, it is copied from c_factors_training.
    nave : int
        Number of epochs in the training set.
    nave_testing : int or None
        Number of epochs in the testing set. If ``None``, it is copied
        from nave.
    """

    def __init__(self, scores_raw, scoring_name, n_classes, c_factors_training,
                 weights, info, times_training, tmin,
                 times_testing=None, c_factors_testing=None, nave=1,
                 nave_testing=None):
        self._scores_raw = scores_raw
        self._scoring_name = scoring_name
        self._nclasses = n_classes
        self._c_factors_training = c_factors_training
        self._weights = weights
        if c_factors_testing is None:
            self._c_factors_testing = self._c_factors_training
        else:
            self._c_factors_testing = c_factors_testing

        self._times_training = times_training
        if times_testing is None:
            self._times_testing = self._times_training
        else:
            self._times_testing = times_testing

        self._info = info
        self._tmin = tmin
        self._nave = nave

        if nave_testing is None:
            nave_testing = nave

        self._nave_testing = nave_testing

    @property
    def info(self):
        """:class:`Measurement info <mne.Info>`"""
        return self._info

    @property
    def tmin(self):
        """float: tmin"""
        return self._tmin

    @property
    def nave(self):
        """int: Number of epochs in the training set."""
        return self._nave

    @property
    def nave_testing(self):
        """int: Number of epochs in the testing set."""
        return self._nave_testing

    @property
    def scores(self):
        """:class:`numpy.ndarray`: The scores of the classification"""
        if self._scores_raw.ndim <= 2:
            return self._scores_raw
        else:
            return numpy.average(self._scores_raw, axis=0)

    @property
    def weights(self):
        """:class:`numpy.ndarray`:The classifier weights"""
        return self._weights

    @property
    def chance_level(self):
        """float: The chance level of the classifier."""
        return 1.0 / self._nclasses

    def diagonal_as_temporal(self):
        """Return the non-generalized results."""
        if len(self._times_training) != len(self._times_testing):
            raise RuntimeError('Cannot show diagonal for non uniform '
                               'dimensions!')
        if self._scores_raw.ndim == 2:
            raw_scores = numpy.diagonal(self._scores_raw)
        else:
            raw_scores = numpy.diagonal(self._scores_raw, axis1=1, axis2=2)

        return TemporalArray(raw_scores=raw_scores,
                             weights=self.weights,
                             n_classes=self._nclasses,
                             info=self.info,
                             tmin=self.tmin,
                             scoring_name=self._scoring_name,
                             nave=self.nave,
                             nave_testing=self.nave_testing,
                             c_factors_training=self._c_factors_training,
                             c_factors_testing=self._c_factors_testing
                             )

    def get_temporal_from_training_interval(self, tmin, tmax):
        """Average the scores of a training interval.

        Parameters
        ----------
        tmin : int
            Start time in seconds of the training interval to average.
        tmax : int
            End time in seconds of the training interval to average.

        Returns
        -------
        data : instance of :class:`TemporalArray`
            The :class:`TemporalArray` with the averaged scores.
        """
        time_idx = numpy.where(
            numpy.logical_and(
                self._times_training >= tmin,
                self._times_training <= tmax))[0]
        if self._scores_raw.ndim == 2:
            raw_scores = numpy.mean(self._scores_raw[time_idx, :], axis=0)
        else:
            raw_scores = numpy.mean(self._scores_raw[:, time_idx, :], axis=1)

        return TemporalArray(raw_scores=raw_scores,
                             weights=self.weights,
                             n_classes=self._nclasses,
                             info=self.info,
                             tmin=self.tmin,
                             scoring_name=self._scoring_name,
                             nave=self.nave,
                             nave_testing=self.nave_testing,
                             c_factors_training=self._c_factors_training,
                             c_factors_testing=self._c_factors_testing
                             )

    def _plot_image(self, values, axes=None, show=True, cmap='Reds',
                    colorbar=True,
                    interpolation='bessel', alpha=None):
        import matplotlib.pyplot as plt

        if not isinstance(axes, plt.Axes):
            fig, axes = plt.subplots(1, 1)
        else:
            fig = axes.get_figure()

        im = axes.imshow(values, interpolation=interpolation, origin='lower',
                         cmap=cmap,
                         extent=numpy.append(self._times_testing[[0, -1]],
                                             self._times_training[[0, -1]]),
                         aspect='auto',
                         alpha=alpha)

        xlabel = 'Testing time (s)'
        ylabel = 'Training time (s)'

        if self._c_factors_testing:
            xlabel = '\n'.join([self._c_factors_testing, xlabel])

        if self._c_factors_training:
            ylabel = '\n'.join([self._c_factors_training, ylabel])

        axes.set_xlabel(xlabel)
        axes.set_ylabel(ylabel)
        axes.set_title('Temporal Generalization')
        axes.axvline(0, color='k')
        axes.axhline(0, color='k')
        old_xlim = axes.get_xlim()
        old_ylim = axes.get_ylim()
        diag = (min(self._times_testing[0], self._times_training[0]),
                max(self._times_testing[-1], self._times_training[-1]))
        axes.plot(diag, diag, color='k')
        axes.set_xlim(old_xlim)
        axes.set_ylim(old_ylim)
        if colorbar:
            plt.colorbar(im, ax=axes)

        if show:
            plt.show()

        return fig

    def plot_scores(self, axes=None, show=True, cmap='Reds', colorbar=True,
                    mask_below_chance=False, interpolation='bessel',
                    insignificant_alpha=0.3):
        """Plot the scores as a Matrix.

        Parameters
        ----------
        axes : :class:`matplotlib.axes.Axes` or None, optional
            The axes where to draw the plot. If ``None``, a new figure is
            created.
        show : bool, optional
            True to actually show the plot.
        cmap : str or :class:`matplotlib.colors.Colormap`, optional
            The colormap.
        colorbar : bool, optional
            Whether to draw the colorbar.
        mask_below_chance : bool, optional
            If True, values below chance level get masked.
        interpolation : str, optional
            The interpolation method used.
        """
        scores = self.scores

        alpha = numpy.ones_like(scores)

        if mask_below_chance:
            mask = scores < self.chance_level
            alpha[mask] = insignificant_alpha

        return self._plot_image(values=scores, axes=axes, show=show, cmap=cmap,
                                colorbar=colorbar, interpolation=interpolation,
                                alpha=alpha)

    def drop_channels(self, chs):
        """Drop the channels from the weights.

        Parameters
        ----------
        chs : list of str
            The names of the channels to drop.
        """
        good_channels_idx = mne.pick_channels(self.info['ch_names'],
                                              include=[],
                                              exclude=chs)

        self._info = mne.pick_info(self.info, good_channels_idx)
        self._weights = self._weights[good_channels_idx, :]

    def crop(self, tmin_training=None, tmax_training=None,
             tmin_testing=None, tmax_testing=None):
        """Crop to times

        """
        if tmin_training is None:
            tmin_training = self._times_training[0]
        if tmax_training is None:
            tmax_training = self._times_training[-1]

        if tmin_testing is None:
            tmin_testing = self._times_testing[0]
        if tmax_testing is None:
            tmax_testing = self._times_testing[-1]

        time_idx_training = numpy.where(
            numpy.logical_and(
                self._times_training >= tmin_training,
                self._times_training <= tmax_training))[0]

        time_idx_testing = numpy.where(
            numpy.logical_and(
                self._times_testing >= tmin_testing,
                self._times_testing <= tmax_testing))[0]

        if self._scores_raw.ndim == 2:
            self._scores_raw = self._scores_raw[time_idx_training, :]
            self._scores_raw = self._scores_raw[:, time_idx_testing]
            if self._weights is not None:
                self._weights = self._weights[time_idx_training, :]
                self._weights = self._weights[:, time_idx_testing]
        else:
            self._scores_raw = self._scores_raw[:, time_idx_training, :]
            self._scores_raw = self._scores_raw[:, :, time_idx_testing]
            if self._weights is not None:
                self._weights = self._weights[:, time_idx_training, :]
                self._weights = self._weights[:, :, time_idx_testing]

        self._times_testing = self._times_testing[time_idx_testing]
        self._times_training = self._times_training[time_idx_training]

        return self


class GeneralizedTemporalFromCollection(GeneralizedTemporalArray):
    """Base class for Temporal Generalization data from multiple subjects.

    The individual elements must match in number of classes, times etc.

    Parameters
    ----------
    data_list : iterable of :class:`GeneralizedTemporalArray`
        A list (or any other iterable) of :class:`GeneralizedTemporalArray`
    raw_scores : :class:`numpy.ndarray` shape (n_channels, n_times) or \
                (n_folds, n_channels, n_times)
        The already processed (i.e., averaged, statistically tested) scores.
    weights : :class:`numpy.ndarray` shape (n_channels, n_times)
        The already processed (i.e., averaged, statistically tested) weights.
    """

    must_be_equal = ['_scoring_name', '_nclasses', '_c_factors_training',
                     '_times_training', 'tmin', '_times_testing',
                     '_c_factors_testing']

    def __init__(self, data_list, raw_scores, weights):
        if not all(isinstance(item, GeneralizedTemporalArray) for item in
                   data_list):
            raise TypeError('Only GeneralizedTemporal instances allowed!')

        self._check_same(data_list)

        first_item = data_list[0]

        nave = first_item.nave
        n_classes = first_item._nclasses
        scoring_name = first_item._scoring_name
        c_factors_training = first_item._c_factors_training
        info = first_item.info
        times_training = first_item._times_training
        tmin = first_item.tmin
        times_testing = first_item._times_testing
        c_factors_testing = first_item._c_factors_testing

        super(GeneralizedTemporalFromCollection, self).__init__(
            scores_raw=raw_scores,
            scoring_name=scoring_name, n_classes=n_classes,
            c_factors_training=c_factors_training, weights=weights, info=info,
            times_training=times_training, tmin=tmin,
            times_testing=times_testing,
            c_factors_testing=c_factors_testing, nave=nave)

    def _check_same(self, data_list):
        first_item = data_list[0]
        for cur_item in data_list:
            for cur_parameter in self.must_be_equal:
                first = getattr(first_item, cur_parameter)
                cur = getattr(cur_item, cur_parameter)
                cmp = first == cur
                if not isinstance(cmp, collections.abc.Iterable):
                    cmp = [cmp]
                if not all(cmp):
                    # raise ValueError('All items must come
                    # from the same contrast!')
                    pass

    def _equalize_channels(self, data_list):
        chan_max_idx = numpy.argmax([c.info['nchan'] for c in data_list])
        chan_template = data_list[chan_max_idx].info['ch_names']

        channels = [set(c.info['ch_names']) for c in data_list]
        common_channels = set(chan_template).intersection(*channels)

        for c in data_list:
            drop_them = list(set(c.info['ch_names']) - common_channels)
            if drop_them:
                c.drop_channels(drop_them)


class GeneralizedTemporalAverage(GeneralizedTemporalFromCollection):
    """Create average instance of Temporal Generalization results.

    Given a list of Temporal Generalization Results, this class computes the
    average of their weights and returns an instance of
    :class:`GeneralizedTemporalArray` so you can plot the average scores
    and weights.

    Parameters
    ----------
    data_list : iterable of :class:`GeneralizedTemporalArray`
        A list (or any other iterable) of :class:`GeneralizedTemporalArray`

    """

    def __init__(self, data_list):
        self._equalize_channels(data_list)

        if None not in [x._weights for x in data_list]:
            weights = numpy.average(numpy.stack([x._weights for x in data_list]),  # noqa
                                    axis=0)
        else:
            weights = None

        scores_raw = numpy.average(
            numpy.stack([x._scores_raw for x in data_list]), axis=0)

        super(GeneralizedTemporalAverage, self).__init__(data_list,
                                                         raw_scores=scores_raw,
                                                         weights=weights)


class GeneralizedTemporalStatistics(GeneralizedTemporalAverage):
    """Run statistics on Temporal Generalization results.

    Given a list of Temporal Generalization Results, this class calculates
    a statistic on the weights and returns an instance of
    :class:`GeneralizedTemporalArray`.
    """

    def __init__(self, data_list, stat_function=scipy.stats.ttest_1samp,
                 popmean=None, **kwargs):
        super(GeneralizedTemporalStatistics, self).__init__(data_list)

        if popmean is None:
            popmean = self.chance_level

        scores_for_stats = numpy.stack(x.scores for x in data_list)

        self.stat_values, self.p = stat_function(scores_for_stats,
                                                 popmean=popmean, **kwargs)
        self.p = self.p / 2.0

    def plot_scores(self, axes=None, show=True, cmap='Reds', colorbar=True,
                    mask_below_chance=False, interpolation='bessel',
                    mask_p=None, insignificant_alpha=0.3):
        """Plot the scores as a Matrix.

        Parameters
        ----------
        axes : :class:`matplotlib.axes.Axes` or None, optional
            The axes where to draw the plot. If ``None``, a new figure is
            created.
        show : bool, optional
            True to actually show the plot.
        cmap : str or :class:`matplotlib.colors.Colormap`, optional
            The colormap.
        colorbar : bool, optional
            Whether to draw the colorbar.
        mask_below_chance : bool, optional
            If True, values below chance level get masked.
        interpolation : str, optional
            The interpolation method used.
        mask_p : float or None, optional
            If set, the plot is masked for the given p-value.
        """
        scores = self.scores

        alpha = numpy.ones_like(scores)

        if mask_below_chance:
            mask = scores < self.chance_level
            alpha[mask] = insignificant_alpha

        if mask_p is not None:
            mask = self.p > mask_p
            alpha[mask] = insignificant_alpha

        return self._plot_image(values=scores, axes=axes, show=show, cmap=cmap,
                                colorbar=colorbar, interpolation=interpolation,
                                alpha=alpha)

    def get_temporal_from_training_interval(self, tmin, tmax):
        """Average the scores of a training interval.

        Parameters
        ----------
        tmin : int
            Start time in seconds of the training interval to average.
        tmax : int
            End time in seconds of the training interval to average.

        Returns
        -------
        data : instance of :class:`TemporalArray`
            The :class:`TemporalArray` with the averaged scores.
        """
        temporal = super(GeneralizedTemporalStatistics,
                         self).get_temporal_from_training_interval(tmin=tmin,
                                                                   tmax=tmax)

        time_idx = numpy.where(
            numpy.logical_and(self._times_training >= tmin,
                              self._times_training <= tmax))[0]

        stats = numpy.mean(self.stat_values[time_idx, :], axis=0)
        temporal.stat_values = stats

        return temporal


class GeneralizedTemporal(GeneralizedTemporalArray):
    """Apply a decoding pipeline using Temporal Generalization.

    Use this class to perform Temporal Generalization decoding. This means
    that a classifier is trained on every sample and then tested on all
    samples. So, if you supply epochs with 100 samples, a classifier gets
    trained on the data data of the first sample. This classifier is then
    tested on the data of the first sample, then the second sample and so on.
    The process is then repeated by training on the second sample and testing
    on all samples and so forth.

    The minimum requirements are the data as :class:`mne.Epochs` and
    the pipeline as :class:`sklearn.pipeline.Pipeline`, which is most
    commonly generated using :func:`sklearn.pipeline.make_pipeline`.

    This class treats every individual ``event_id`` as an individual target.
    If you want to combine multiple ``events_id`` into one target, you can
    use :func:`epochs.collapse_conditions`.

    If you want the training and testing data to be different, you can supply
    the training data to ``epochs`` and the testing data to ``epochs_test``.

    Cross validation will be run only if training and testing data are equal
    (i.e., when ``epochs_test=None``).

    Parameters
    ----------
    epochs : :class:`mne.Epochs`
        The epochs to apply the classifier pipeline to.
    pipeline : :class:`sklearn.pipeline.Pipeline`
        The classifier pipeline to use. Most likely created with
        :func:`sklearn.pipeline.make_pipeline`
    epochs_test : :class:`mne.Epochs` or None, optional
        If set, the classifier gets tested on this data. The event_ids must
        be equal. No crossvalidation will be run in this case.
    cv : int, optional
        The amount of folds for cross validation
    scoring : str or callable, optional
        The scoring function to use.
    n_jobs : int, optional
        Number of CPU cores to use

    """

    def __init__(self, epochs, pipeline, epochs_test=None, cv=2,
                 scoring='accuracy', n_jobs=-1, metadata_querylist=None,
                 metadata_querylist_test=None, verbose=True):
        times_training = epochs.times
        times_testing = None
        nave = len(epochs)
        nave_testing = None

        new_info = mne.create_info(epochs.info['ch_names'],
                                   epochs.info['sfreq'])

        with new_info._unlock():
            new_info['chs'] = copy.deepcopy(epochs.info['chs'])

        self._decoder = mne.decoding.GeneralizingEstimator(pipeline,
                                                           n_jobs=n_jobs,
                                                           scoring=scoring,
                                                           verbose=verbose)
        X = epochs.get_data()
        if metadata_querylist is None:
            y = epochs.events[:, 2]
        else:
            y = _get_Y_from_metadata(epochs, metadata_querylist)
            if metadata_querylist_test is None:
                metadata_querylist_test = metadata_querylist

        n_classes = numpy.unique(y).size

        c_factors_training = ''
        if 'collapsed_factors' in epochs.info:
            c_factors_training = epochs.info['collapsed_factors']

        c_factors_testing = None

        if isinstance(epochs_test, mne.BaseEpochs):
            logging.info(
                'Using epochs argument for training and epochs_test argument '
                'for testing')

            X_test = epochs_test.get_data()
            if metadata_querylist_test is None:
                y_test = epochs_test.events[:, 2]
            else:
                y_test = _get_Y_from_metadata(epochs_test,
                                              metadata_querylist_test)
            nave_testing = len(epochs_test)

            if 'collapsed_factors' in epochs_test.info:
                c_factors_testing = epochs_test.info['collapsed_factors']

            times_testing = epochs_test.times

            if numpy.setdiff1d(y, y_test).size != 0:
                raise ValueError(
                    'Training and testing set must have the same targets!')

            if X.shape[1] != X_test.shape[1]:
                raise ValueError('Training and testing set must have the '
                                 'same number of channels!')

            self._decoder.fit(X, y)
            scores_raw = self._decoder.score(X_test, y_test)

        else:
            scores_raw = mne.decoding.cross_val_multiscore(self._decoder, X, y,
                                                           cv=cv)

        try:
            self._decoder.fit(X, y)
            weights = mne.decoding.get_coef(self._decoder, 'patterns_',
                                            inverse_transform=True)
            if weights.ndim == 3:
                weights = numpy.average(weights, axis=1)
        except ValueError:
            weights = None

        super(GeneralizedTemporal, self).__init__(scores_raw=scores_raw,
                                                  scoring_name=scoring,
                                                  n_classes=n_classes,
                                                  c_factors_training=c_factors_training,  # noqa
                                                  c_factors_testing=c_factors_testing,  # noqa
                                                  times_training=times_training,  # noqa
                                                  times_testing=times_testing,
                                                  weights=weights,
                                                  tmin=epochs.times[0],
                                                  info=new_info,
                                                  nave=nave,
                                                  nave_testing=nave_testing)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import mne
import numpy as np


def filter_event_id(event_id, conditions):
    """Filter mne-python style event_ids.

    Parameters
    ----------
    event_id : dict
        The event_id
    conditions : str or list or tuple
        The conditions to keep

    Returns
    -------
    The filtered event_id

    """
    conditions = [conditions] if not isinstance(conditions,
                                                (list, tuple)) else conditions
    good_event_keys = mne.epochs._hid_match(event_id, conditions)

    new_event_id = {key: value for key, value in event_id.items() if
                    key in good_event_keys}
    return new_event_id


def read_events_from_analogue(raw, tolerance=1, trigger_channels=None):
    if trigger_channels is None:
        trigger_channels = ['STI001', 'STI002', 'STI003',
                            'STI004', 'STI005', 'STI006', 'STI007', 'STI008']

    trigger_data = raw[mne.pick_channels(raw.ch_names, trigger_channels)][0] / 5  # noqa

    bit_mult = 2 ** np.arange(0, trigger_data.shape[0])

    trigger_values = np.sum((trigger_data.T * bit_mult).T, axis=0).astype(int)

    trigger_idx = np.where(np.diff(trigger_values) > 0)[0] + 1
    tmp_events = np.zeros((trigger_idx.shape[0], 3))
    tmp_events[:, 0] = trigger_idx + raw.first_samp
    tmp_events[:, 2] = trigger_values[trigger_idx]

    bad_triggers_idx = np.where(np.diff(tmp_events[:, 0]) <= tolerance)[0]

    tmp_events[bad_triggers_idx + 1, 0] = tmp_events[bad_triggers_idx, 0]
    events = np.delete(tmp_events, bad_triggers_idx, axis=0)

    return events.astype(int)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

import mne
import copy
import numpy
import logging

from .helpers import _get_Y_from_metadata


class TemporalArray(mne.EvokedArray):
    """Base class for temporal decoding.

    Parameters
    ----------
    raw_scores : :class:`numpy.ndarray` shape (n_times) or (n_folds, n_times)
        The scores of the classification
    weights : :class:`numpy.ndarray` shape (n_channels, n_times)
        Classifier weights
    n_classes : int
        The number of classes
    info : dict
        Info dict
    tmin : float
        Time of the first sample in seconds
    scoring_name : str
        Name of the scoring function (i.e. Accuracy...)
    c_factors_training : str
        Name of the factors over which was collapsed in the training set
    nave : int, optional
        Number of epochs in the training set.
    nave_testing : int or None
        Number of epochs in the testing set. If ``None``, it is copied
        from nave.
    c_factors_testing : None or str
        Name of the factors over which was collapsed in the testing set.
        If ``None``, it is copied from c_factors_training.

    """

    def __init__(self, raw_scores, weights, n_classes, info, tmin,
                 scoring_name,
                 c_factors_training, nave=1, nave_testing=None,
                 c_factors_testing=None):
        self._scores_raw = raw_scores
        self._nclasses = n_classes
        self._scoring_name = scoring_name
        self._c_factors_training = c_factors_training
        if c_factors_testing is None:
            self._c_factors_testing = self._c_factors_training
        else:
            self._c_factors_testing = c_factors_testing

        if nave_testing is None:
            nave_testing = nave

        self.nave_testing = nave_testing
        self.nave = nave

        super(TemporalArray, self).__init__(weights, info, tmin, nave=nave)

    @property
    def scores(self):
        """:class:`numpy.ndarray`: The scores of the classification"""
        if self._scores_raw.ndim == 1:
            return self._scores_raw
        else:
            return numpy.average(self._scores_raw, axis=0)

    @property
    def nclasses(self):
        return self._nclasses

    @property
    def chance_level(self):
        """float: The chance level of the classifier."""
        return 1.0 / self._nclasses

    def plot_scores(self, axes=None, show=True):
        """Plot the scores as a line plot.

        Parameters
        ----------
        axes : :class:`matplotlib.axes.Axes` or None, optional
            The axes where to draw the plot. If ``None``, a new figure is
            created.
        show : bool, optional
            True to actually show the plot.
        """
        import matplotlib.pyplot as plt

        if not isinstance(axes, plt.Axes):
            fig, axes = plt.subplots(1, 1)
        else:
            fig = axes.get_figure()

        axes.plot(self.times, self.scores, label='score')
        axes.axhline(self.chance_level, color='k', linestyle='--',
                     label='chance')
        axes.set_xlabel('Times')
        axes.set_ylabel(self._scoring_name)
        axes.legend()
        axes.axvline(.0, color='k', linestyle='-')
        axes.set_title('Sensor space decoding')
        if show:
            plt.show()

        return fig


class Temporal(TemporalArray):
    """Apply a decoding pipeline to every sample.

    Use this class to perform Temporal decoding. This means that a classifier
    is trained on every sample and tested on that very sample.

    The minimum requirements are the data as :class:`mne.Epochs` and
    the pipeline as :class:`sklearn.pipeline.Pipeline`, which is most
    commonly generated using :func:`sklearn.pipeline.make_pipeline`.

    This class treats every individual ``event_id`` as an individual target.
    If you want to combine multiple ``events_id`` into one target, you can
    use :func:`epochs.collapse_conditions`.

    If you want the training and testing data to be different, you can supply
    the training data to ``epochs`` and the testing data to ``epochs_test``.

    Cross validation will be run only if training and testing data are equal
    (i.e., when ``epochs_test=None``).

    Parameters
    ----------
    epochs : :class:`mne.Epochs`
        The epochs to apply the classifier pipeline to.
    pipeline : :class:`sklearn.pipeline.Pipeline`
        The classifier pipeline to use. Most likely created with
        :func:`sklearn.pipeline.make_pipeline`
    epochs_test : :class:`mne.Epochs` or None, optional
        If set, the classifier gets tested on this data. The event_ids must
        be equal. No crossvalidation will be run in this case.
    cv : int, optional
        The amount of folds for cross validation
    scoring : str or callable, optional
        The scoring function to use.
    n_jobs : int, optional
        Number of CPU cores to use
    """

    def __init__(self, epochs, pipeline, epochs_test=None, cv=2,
                 scoring='accuracy', n_jobs=-1, metadata_querylist=None,
                 metadata_querylist_test=None):

        new_info = mne.create_info(epochs.info['ch_names'],
                                   epochs.info['sfreq'])
        new_info['chs'] = copy.deepcopy(epochs.info['chs'])
        self._decoder = mne.decoding.SlidingEstimator(pipeline, n_jobs=n_jobs,
                                                      scoring=scoring)
        X = epochs.get_data()
        if metadata_querylist is None:
            y = epochs.events[:, 2]
        else:
            y = _get_Y_from_metadata(epochs, metadata_querylist)
            if metadata_querylist_test is None:
                metadata_querylist_test = metadata_querylist

        nave = len(epochs)
        nave_testing = None

        n_classes = numpy.unique(y).size

        c_factors_training = ''
        if 'collapsed_factors' in epochs.info:
            c_factors_training = epochs.info['collapsed_factors']

        c_factors_testing = None

        if isinstance(epochs_test, mne.BaseEpochs):
            logging.info('Using epochs argument for training and epochs_test '
                         'argument for testing')
            X_test = epochs_test.get_data()
            if metadata_querylist is None:
                y_test = epochs_test.events[:, 2]
            else:
                y_test = _get_Y_from_metadata(epochs_test,
                                              metadata_querylist_test)
            nave_testing = len(epochs_test)

            if 'collapsed_factors' in epochs_test.info:
                c_factors_testing = epochs_test.info['collapsed_factors']

            if numpy.setdiff1d(y, y_test).size != 0:
                raise ValueError(
                    'Training and testing set must have the same targets!')

            if X.shape[1:] != X_test.shape[1:]:
                raise ValueError(
                    'Training and testing set must have the same number '
                    'of channels and samples!')

            self._decoder.fit(X, y)
            scores_raw = self._decoder.score(X_test, y_test)
        else:
            scores_raw = mne.decoding.cross_val_multiscore(self._decoder, X, y,
                                                           cv=cv)
            self._decoder.fit(X, y)

        weights = mne.decoding.get_coef(self._decoder, 'patterns_',
                                        inverse_transform=True)
        if weights.ndim == 3:
            weights = numpy.average(weights, axis=1)

        super(Temporal, self).__init__(raw_scores=scores_raw,
                                       weights=weights,
                                       n_classes=n_classes,
                                       info=new_info,
                                       tmin=epochs.times[0],
                                       scoring_name=scoring,
                                       nave=nave,
                                       nave_testing=nave_testing,
                                       c_factors_testing=c_factors_testing,
                                       c_factors_training=c_factors_training)

    @property
    def decoder(self):
        """The decoder."""
        return self._decoder



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.

from codecs import open

import os.path
from setuptools import setup, find_packages

# find the location of this file
this_directory = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the Module Version from the VERSION file
with open(os.path.join(this_directory, 'VERSION'), encoding='utf-8') as f:
    version = f.read()

# define required modules
required = ['mne', 'scipy', 'numpy', 'obob_condor', 'pandas']

setup(
    name='obob_mne',
    packages=find_packages(),
    url='https://gitlab.com/obob/obob_mne',
    version=version,
    description='I am obob_mne',
    long_description=long_description,
    license='GPL3',
    author='Thomas Hartmann',
    author_email='thomas.hartmann@th-ht.de',
    install_requires=required,
    entry_points={
                'console_scripts': [
                        'import_subject_to_freesurfer=obob_mne.mri.cmd.import_subject:import_subject',
                        'make_freesurfer_bem=obob_mne.mri.cmd.make_freesurfer_bem:make_freesurfer_bem_cmd',
                        'make_freesurfer_sourcespaces=obob_mne.mri.cmd.make_source_space:make_freesurfer_sourcespaces_cmd'
                        ]
                }
)



# -*- coding: UTF-8 -*-
# Copyright (c) 2018, Thomas Hartmann
#
# This file is part of the obob_mne Project, see:
# https://gitlab.com/obob/obob_mne
#
#    obob_mne is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    obob_mne is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with obob_subjectdb. If not, see <http://www.gnu.org/licenses/>.
from copy import deepcopy

import mne
import os
import logging
import glob
import numpy as np
import re
import obob_mne.events
import pandas as pd
import mne.io.fiff


class LoadFromSinuhe(mne.io.fiff.Raw):
    """:class:`mne.io.Raw` mixin to facilitate loading data from sinuhe.

    By including this mixin in your study specific ``Raw`` class, you can
    facilitate loading the raw data files.

    Supposed your fif files follow the usual pattern: '19800908igdb_run01.fif',
    you can define a ``Raw`` class like this one:

    .. code-block:: python

        class Raw(mne.io.fiff.Raw, LoadFromSinuhe):
            study_acronym = 'test_study'

    Loading run 2 of subject '19800908igdb' can then be done like this:

    .. code-block:: python

        raw_data = Raw(subject_id='19800908igdb', block_nr=2, preload=True)

    """

    sinuhe_root = '/mnt/sinuhe/data_raw'
    study_acronym = None
    file_glob_patterns = [
        '%s_run%02d.fif',
        '%s_block%02d.fif',
        '%s_run%d.fif',
        '%s_block%d.fif'
    ]

    excluded_subjects = None
    replace_subject_ids = dict()

    subject_id_regex = r'^[1-2][0-9]{7}[a-z]{4}$'

    def __init__(self, subject_id, block_nr=None, **kwargs):
        if not self.study_acronym:
            raise ValueError('You must set the study_acronym parameter!')

        if not os.path.isdir(
                os.path.join(self.sinuhe_root, self.study_acronym)):
            raise ValueError('Cannot find your study on sinuhe!')

        if os.path.isfile(subject_id):
            logging.info('A filename was submitted. Loading...')
            final_fname = subject_id
        else:
            final_fname = self.get_fif_filename(subject_id, block_nr)

        super(LoadFromSinuhe, self).__init__(final_fname, **kwargs)

    @classmethod
    def get_all_subjects(cls):
        """Return a list of all subjects in the study.

        Returns
        -------
        all_subjects : list
            A list of strings with all subjects codes found.
        """
        if not cls.study_acronym:
            raise ValueError('You must set the study_acronym parameter!')

        if not os.path.isdir(os.path.join(cls.sinuhe_root, cls.study_acronym)):
            raise ValueError('Cannot find your study on sinuhe!')

        study_folder = os.path.join(cls.sinuhe_root, cls.study_acronym)
        all_fif_files = glob.glob(
            os.path.join(study_folder, '*', '*', '*.fif'))
        all_fif_files = [os.path.basename(x) for x in all_fif_files]
        all_subjects = set([x[0:12] for x in all_fif_files if
                            re.match(cls.subject_id_regex, x[0:12])])

        if cls.excluded_subjects:
            all_subjects -= set(cls.excluded_subjects)

        if cls.replace_subject_ids:
            all_subjects -= set(cls.replace_subject_ids.keys())

        return all_subjects

    @classmethod
    def get_number_of_runs(cls, subject_id):
        """Return the number of runs for the given subject.

        Parameters
        ----------
        subject_id : str
            The subject id

        Returns
        -------
        n_runs : int
            The number of runs for the subject.
        """
        n_runs = 1
        while cls.get_fif_filename(subject_id, n_runs) is not None:
            n_runs += 1

        return n_runs - 1

    @classmethod
    def get_fif_filename(cls, subject_id, run_nr):
        """Find the fif file for the subject and run.

        Parameters
        ----------
        subject_id : str
            The subject id
        run_nr : int
            The run number

        Returns
        -------
        fname : str
            The filename of the respective fif file

        """
        if not cls.study_acronym:
            raise ValueError('You must set the study_acronym parameter!')

        if not os.path.isdir(os.path.join(cls.sinuhe_root, cls.study_acronym)):
            raise ValueError('Cannot find your study on sinuhe!')

        study_folder = os.path.join(cls.sinuhe_root, cls.study_acronym)

        all_subject_ids = (subject_id.lower(), subject_id.upper())

        if cls.replace_subject_ids:
            if set(all_subject_ids) & set(cls.replace_subject_ids.values()):
                tmp_all_ids = all_subject_ids
                for cur_subject_id in set(tmp_all_ids) & set(
                        cls.replace_subject_ids.values()):
                    idx = list(cls.replace_subject_ids.values()).index(
                        cur_subject_id)
                    all_subject_ids += (list(
                        cls.replace_subject_ids.keys())[idx],)

        fname = None
        for cur_subject_id in all_subject_ids:
            if cls.excluded_subjects:
                if cur_subject_id in cls.excluded_subjects:
                    raise ValueError(
                        'Subject %s is excluded!' % (cur_subject_id,))

            for cur_glob_pattern in cls.file_glob_patterns:
                tmp_result = glob.glob(
                    os.path.join(
                        study_folder, '*', '*',
                        cls.get_glob_pattern(
                            cur_glob_pattern, cur_subject_id, run_nr
                        )
                    )
                )
                if tmp_result:
                    fname = tmp_result[-1]

        return fname

    @classmethod
    def get_glob_pattern(cls, cur_glob_pattern, cur_subject_id, run_nr):
        if run_nr is not None:
            return cur_glob_pattern % (cur_subject_id, run_nr)
        else:
            return cur_glob_pattern % (cur_subject_id, )


class AdvancedEvents(mne.io.fiff.Raw):
    """Integrate event loading and handling into :class:`mne.io.Raw`.

    Including this mixin in your study specific ``Raw`` class provides event
    handling features directly in that class.

    More specifically, it provides three extra properties:

    1. events
    2. event_id
    3. evt_metadata

    Which are automatically filled and kept up-to-date. They correspond to
    the respective meaning in :class:`mne.Epochs`.

    You can also create a subclass of this class and use
    :meth:`_process_events` to process the events (fill the event_id,
    modify the event codes....)
    """

    trigger_min_duration = 9e-3

    def __init__(self, *args, **kwargs):
        self._events = None
        self._event_id = None
        self._evt_metadata = None

        super(AdvancedEvents, self).__init__(*args, **kwargs)
        self._load_events()

    def _load_events(self):
        self._event_id = dict()
        self._events = mne.find_events(self,
                                       min_duration=self.trigger_min_duration)

        self._process_events()

    def _process_events(self):
        pass

    def get_filtered_event_id(self, condition_filter):
        """Return a filtered version of the event_id field.

        Refer to :meth:`obob_mne.events.filter_event_id` for further
        details.
        """
        return obob_mne.events.filter_event_id(self.event_id, condition_filter)

    def has_filtered_events(self, condition_filter):
        """Check whether the event_ids are present.

        Parameters
        ----------
        condition_filter : str
            The event_ids to check

        Returns
        -------
        has_events : bool
            True if the filtered events are present.

        """
        try:
            self.get_filtered_event_id(condition_filter)
        except KeyError:
            return False

        return len(self.get_filtered_event_id(condition_filter)) > 0

    def resample(self, *args, **kwargs):
        """Resample the data and reloads the events.

        For the rest, refer to :meth:`mne.io.Raw.resample`.

        """
        self._events = None
        self._event_id = None
        self._evt_metadata = None

        super(AdvancedEvents, self).resample(*args, **kwargs)
        self._load_events()

    @property
    def events(self):
        """:class:`numpy.ndarray`: The event matrix."""
        if not isinstance(self._events, np.ndarray):
            self._load_events()

        return self._events

    @property
    def event_id(self):
        """dict: The event_ids"""
        if not isinstance(self._events, np.ndarray):
            self._load_events()

        return self._event_id

    @property
    def evt_metadata(self):
        """:class:`pandas.DataFrame`: The metadata"""
        if not isinstance(self._events, np.ndarray):
            self._load_events()

        return self._evt_metadata


class AutomaticBinaryEvents(AdvancedEvents):
    """Mixin for binary events.

    If your triggers code events with binary triggers, this mixin can help you
    a lot.

    Let's suppose, you have an experiment with two types of blocks.
    At the beginning of each block, the type of the block is signalled by a
    a trigger code:

    1. Attend Auditory: Trigger 1
    2. Attend Visual: Trigger 2

    Then you present either:

    1. A tone: Trigger 4
    2. An image: Trigger 8

    And sometimes, one of them is an oddball which is marked by adding 1 to
    the trigger codes.

    In this case, you can use this mixin and write something like this:

    .. code-block:: python

        class Raw(mne.io.fiff.Raw, LoadFromSinuhe, AutomaticBinaryEvents):
            study_acronym = 'test_study'

            condition_triggers = {
                'attention': {
                    'auditory': 1,
                    'visual': 2
                }
            }

            stimulus_triggers = {
                'modality': {
                    'audio': 4,
                    'visual': 8
                },
                'oddball': 1
            }

    This will automatically result in mne-python aware event_ids like:

    ``'attention:visual/modality:audio/oddball:True'``

    """

    condition_triggers = None
    stimulus_triggers = None

    def __init__(self, *args, **kwargs):

        if not isinstance(self.stimulus_triggers, dict):
            raise ValueError('Please set stimulus_triggers to a dictionary')

        super(AutomaticBinaryEvents, self).__init__(*args, **kwargs)

    def _decode_bin_trigger(self, trigger_value, trigger_dict_item):
        if isinstance(trigger_dict_item, dict):
            for key, value in trigger_dict_item.items():
                if trigger_value & value:
                    return key
        else:
            if trigger_value & trigger_dict_item:
                return 'yes'
            else:
                return 'no'

    def _process_events(self):
        super(AutomaticBinaryEvents, self)._process_events()

        conditions_string_list = list()
        conditions_metadata = {}
        all_raw_metadata = []

        if self.condition_triggers:
            condition_trigger = self._events[0, 2]
            self._events = np.delete(self._events, (0), axis=0)

            for allcond_key, allcond_value in self.condition_triggers.items():
                decoded_trigger = self._decode_bin_trigger(condition_trigger,
                                                           allcond_value)
                conditions_string_list.append(
                    '%s:%s' % (allcond_key, decoded_trigger))
                conditions_metadata[allcond_key] = decoded_trigger

        all_event_codes = np.unique(self._events[:, 2])
        original_events = self._events.copy()

        for cur_code in all_event_codes:
            cur_string_list = list(conditions_string_list)

            for stim_group_name, stim_group_choices in \
                    self.stimulus_triggers.items():
                cur_string_list.append('%s:%s' % (stim_group_name,
                                                  self._decode_bin_trigger(
                                                      cur_code,
                                                      stim_group_choices)))

            new_event_key = '/'.join(cur_string_list)
            new_event_code = np.mod(np.abs(hash(new_event_key)), 4000)
            self._event_id[new_event_key] = new_event_code
            self._events[self._events[:, 2] == cur_code, 2] = new_event_code

        for cur_evt_code in original_events[:, 2]:
            cur_metadata = deepcopy(conditions_metadata)
            for stim_group_name, stim_group_choices in \
                    self.stimulus_triggers.items():
                cur_metadata[stim_group_name] = self._decode_bin_trigger(
                    cur_evt_code, stim_group_choices
                )

            all_raw_metadata.append(cur_metadata)

        self._evt_metadata = pd.DataFrame(all_raw_metadata)


class AutomaticBinaryEventsWithMetadata(AutomaticBinaryEvents):
    """Mixin for binary events with metadata.

    """
    def _process_events(self):
        super(AutomaticBinaryEvents, self)._process_events()

        conditions_metadata = {}
        all_raw_metadata = []

        if self.condition_triggers:
            condition_trigger = self._events[0, 2]
            self._events = np.delete(self._events, (0), axis=0)

            for allcond_key, allcond_value in self.condition_triggers.items():
                conditions_metadata[allcond_key] = self._decode_bin_trigger(
                    condition_trigger,
                    allcond_value
                )

        for cur_evt_code in self._events[:, 2]:
            cur_metadata = deepcopy(conditions_metadata)
            for stim_group_name, stim_group_choices in \
                    self.stimulus_triggers.items():
                cur_metadata[stim_group_name] = self._decode_bin_trigger(
                    cur_evt_code, stim_group_choices
                )

            all_raw_metadata.append(cur_metadata)

        self._evt_metadata = pd.DataFrame(all_raw_metadata)
